/*
ЧПП 0.5
автор lostbyte32@mail.com
ограничения - согласно j4l последней редакции, ЧПП само собой.
*/
class Engine {

    static mainLoop(timeStamp) {

        const engine = window.mycanvas;
        //console.log(timeStamp);
        if (engine) {
            engine.update();
            engine.draw(timeStamp);
        }
        // Keep requesting new frames
        window.requestAnimationFrame(Engine.mainLoop);
    }

    constructor(witdh, height, canvas_id = null, with_render_update = false, with_audio = true, asHtml5 = false) {
        this.html5 = false;
        this.sceneManager = new SceneManager(this);
        this.images = [];
        this.renderObjects = [];
        this.displayList = [];
        this.canvas = document.body;
        this.canvas.style.width = witdh;
        this.canvas.style.height = height;
        this.texturePacks = [];
        this.lastUpdate = 0;
        this.lastFPS = 0;
        this.currentTime = 0;
        this.runFromStart = 0;
        this.input = new InputMgr(this);
        const engine = this;
        this.preloadMgr = new Preloader();
        this.updateTimer = null;
        if (with_render_update) {
            this.startRendererUpdate();
        }
        if (with_audio) {
            this.audioEngine = new AudioEngine(this);
        }
        if (canvas_id && asHtml5) {
            this.canvas = document.getElementById(canvas_id);
            this.context = this.canvas.getContext('2d');
            this.html5 = true;
            window.requestAnimationFrame(Engine.mainLoop);
            window.mycanvas = this;
        }
        else {
            this.canvas = document.getElementById(canvas_id);
            const rect = this.canvas.getBoundingClientRect();
            this.canvas.width = parseInt(rect.width);
            this.canvas.height = parseInt(rect.height);
        }


    }

    checkFps(timeStamp) {
        this.runFromStart += this.currentTime;

        this.currentTime = (timeStamp - this.lastUpdate) / 1000;
        this.lastUpdate = timeStamp;
        //console.log(this.lastUpdate);
        // Calculate fps
        this.lastFPS = Math.round(1 / this.currentTime);
        //console.log(this.lastFPS);
        return this.lastFPS;

    }

    detectEdgeCollisions() {
        let obj;
        for (let i = 0; i < this.gameObjects.length; i++) {
            obj = this.gameObjects[i];

            // Check for left and right
            if (obj.x < obj.radius) {
                obj.vx = Math.abs(obj.vx) * restitution;
                obj.x = obj.radius;
            } else if (obj.x > this.canvas.width - obj.radius) {
                obj.vx = -Math.abs(obj.vx) * restitution;
                obj.x = canvasWidth - obj.radius;
            }

            // Check for bottom and top
            if (obj.y < obj.radius) {
                obj.vy = Math.abs(obj.vy) * restitution;
                obj.y = obj.radius;
            } else if (obj.y > canvasHeight - obj.radius) {
                obj.vy = -Math.abs(obj.vy) * restitution;
                obj.y = canvasHeight - obj.radius;
            }
        }
    }
    draw(timeStamp) {

        this.context.clearRect(0, 0, this.canvas.width, this.canvas.height);
        let fps = this.checkFps(timeStamp);
        const rect = new Rectangle(this.context, 0, 0, 200, 100, 'black');
        const text = new TextElement(this.context, 10, 30, fps);
        this.context.fillStyle = '#ff8080';
        let rectX = 0;
        let rectY = 0;
        let movingSpeed = 50;
        rect.draw();
        // Draw a rectangle on the canvas
        //this.moveToXY(rect, 300, 0, 10);
        //rectX = this.easeInOutQuint(this.runFromStart, 50, 500, 1.5);
        //rectY = this.easeLinear(this.runFromStart, 50, 250, 1.5);
    }


    update() {

    }



    startRendererUpdate(delta = 22) {
        const fps = delta * 1001;
        const time = Helper.get_time();
        this.updateTimer = time;

        setInterval(() => {
            requestAnimationFrame(() => {
                engine.update()
            });
        }, delta / 30);
    }

    renderNow(deelay = 0.0) {
        if (deelay === 0.0) {
            this.update();
        } else
            this.laterUpdate(deelay);
    }

    getImageByName(name) {
        let found = null;
        this.images.forEach((img) => {
            if (img.name === name)
                found = img;
        });
        return found;
    }

    laterUpdate(deelay) {
        const engine = this;
        const timerOut = setTimeout(() => {
            engine.update();
        }, deelay)
    }

    old_update() {
        if (this.sceneManager.getScenes().length > 0) {
            this.sceneManager.updateVisibleScenes();
        }
    }

    loadTexturePack(jsonUrl, callback) {
        const texture = new TexturePack(jsonUrl);
        const engine = this;
        let images = [];
        texture.loadJSONData((err, data) => {
            if (!err && data) {
                texture.loadImages(engine.preloadMgr, (err, info) => {
                    if (!err) {
                        for (const img of info) {
                            images.push(img);
                        }
                        engine.preloadMgr.images = images;

                        return callback(null, engine.preloadMgr.images);
                    }
                    else {
                        return callback(new Error('Error loading texture assets'));
                    }
                });
            } else {
                return callback(new Error('Error loading texture assets'));
            }
        });
    }

    addToRender(ready_sprite, x, y) {
        const engine = this;
        const obj = ready_sprite;
        ready_sprite = ready_sprite.img;
        if (this.canvas) {
            ready_sprite.style.position = 'absolute';
            ready_sprite.style.top = y;
            ready_sprite.style.left = x;
            engine.canvas.appendChild(ready_sprite);
            this.addToCommonRender(obj);
            this.updateCommonRenderer();
        }
    }

    addToCommonRender(renderObject) {
        this.renderObjects.push(renderObject);
    }

    render(objectsForRender) {
        const engine = this;
        requestAnimationFrame(() => { engine.clearRenderer() });
        objectsForRender.forEach((object) => {
            if (object && object.visible) {
                const x = parseInt(object.style.left);
                const y = parseInt(object.style.top);
                object.pos_x = x;
                object.pos_y = y;

                if (engine.html5 && engine.context) {
                    engine.context.drawImage(object, x, y, 100, 100);
                    this.updateCommonRenderer(true);
                } else {
                }

            }
        });
    }

    createAnimationForSprite(sprite, frames = [], speed = 300, loop = true) {
        let anim = [];
        const engine = this;
        sprite.img.onload = function () {

            const spriteImage = sprite.img;
            const spriteWidth = parseInt(spriteImage.style.nw);
            const spriteHeigth = parseInt(spriteImage.style.nh);
            frames.forEach((frame) => {
                const isLoaded = engine.preloadMgr.getImageByName(frame, (err, data) => {
                    if (!err) {
                        let newSrpite = data;
                        if (spriteWidth > 0 || spriteHeigth > 0) {
                            newSrpite = ImageConverter.scaleImage(data, spriteWidth, spriteHeigth);
                            newSrpite.img.style.nw = spriteWidth;
                            newSrpite.img.style.nh = spriteHeigth;
                        }
                        newSrpite.img.style.position = sprite.img.style.position;
                        newSrpite.img.style.top = sprite.img.style.top;
                        newSrpite.img.style.left = sprite.img.style.left;
                        newSrpite.img.style.filter = sprite.img.style.filter;
                        anim.push(newSrpite);
                        engine.addToCommonRender(newSrpite);
                    } else {
                        console.log(err);
                    }

                });
            });
            if (anim.length > 0) {
                sprite.visible = false;
                const animator = new Animator(sprite, engine, anim, true, 300);
                sprite.addAnimation(animator);
            }
        };
    }

    updateCommonRenderer(force = false) {
        const engine = this;

        this.clearRenderer();

        requestAnimationFrame(() => {
            if (this.renderObjects.length) {

                this.renderObjects.forEach((renderObject) => {

                    if (renderObject.visible && renderObject.isNeedUpdate()) {
                        engine.canvas.appendChild(renderObject.getImage());
                        renderObject.update();
                    }

                });

            }
        });
    }

    addSprite(img_name, url = null, pos_x = 0, pos_y = 0, scale = {}, callback) {
        let resize = false;
        if (scale && scale.width || scale.height) {
            resize = true;
        }
        const engine = this;
        let isLoaded = null;
        if (this.preloadMgr) {
            isLoaded = this.preloadMgr.getImageByName(img_name, (err, data) => {

                if (!err) {
                    if (data) {
                        let newImg = data.img.cloneNode(true);
                        if (resize) {
                            newImg = ImageConverter.scaleImage(data, scale.width, scale.height).getImage();
                            newImg.style.nw = scale.width;
                            newImg.style.nh = scale.height;
                        }
                        const imgObject = new Img(data.name, data.url, pos_x, pos_y);

                        newImg.style.position = 'absolute';
                        newImg.style.top = pos_y + 'px';
                        newImg.style.left = pos_x + 'px';
                        newImg.style.maxWidth = '100%';
                        newImg.style.objectFit = 'scale-down';
                        newImg.style.filter = "drop-shadow(16px 16px 10px black)";
                        engine.canvas.appendChild(newImg);
                        imgObject.img = newImg;
                        engine.addToCommonRender(imgObject);
                        engine.images.push(imgObject);
                        callback(null, imgObject);
                    }

                }
                else {
                    engine.tryGetWithoutCache(img_name, url, pos_x, pos_y, scale, (err, img) => {
                        if (!err) {
                            callback(null, img);
                        }
                        else {
                            callback(new Error('nope'));
                        }
                    });

                }

            });


        }
    }

    tryGetWithoutCache(img_name, url = null, pos_x = 0, pos_y = 0, scale = {}, callback) {
        const engine = this;
        let resize;
        if (scale && scale.width || scale.height) {
            resize = true;
        }

        const img = engine.preloadMgr.preloadImage(img_name, url, { x: pos_x, y: pos_y }, (img) => {
            let newImg;
            if (img.img) {
                newImg = JSON.parse(JSON.stringify(img));
                newImg.img = new Image(img.img);
                const z = engine.displayList.length + 1;

                if (!engine.html5) {
                    if (engine.canvas) {
                        newImg = engine.canvas.appendChild(img.img);
                        newImg.style.position = 'absolute';
                        newImg.style.top = /*engine.canvas.height / 2 -*/ pos_y + 'px';
                        newImg.style.left = /*engine.canvas.width / 2 -*/ (pos_x - engine.canvas.width / 2) + 'px';
                        newImg.style.zIndex = z;
                        newImg.style.objectFit = 'scale-down';
                        if (resize) {
                            newImg.width = scale.width;
                            newImg.height = scale.height;
                        }
                        engine.canvas.appendChild(newImg);
                    }
                }
                engine.images.push(newImg);
                engine.displayList.push(newImg);
                callback(null, newImg);
            }
            else {
                callback(new Error('Error something at point 1'));
            }
        });
    }

    clearRenderer(doneCallback) {
        if (!this.html5 && this.canvas) {
            const engine = this;
            if (this.renderObjects.length) {

                this.renderObjects.forEach((renderObject) => {
                    if (renderObject.isNeedUpdate() || !renderObject.visible) {
                        renderObject.img.remove();
                    }
                })

            }
        } else if (this.html5) {
            if (this.context && this.canvas) {
                this.context.clearRect(0, 0, this.canvas.width, this.canvas.height);
            }
        }
    }

    addTextLayer(text, layerId, pos_x, pos_y) {
        let textElem = document.createElement('div');
        textElem.innerHTML = text;
        textElem.style.position = 'absolute';
        textElem.style.left = pos_x + 'px';
        textElem.style.top = pos_y + 'px';
        if (this.canvas)
            this.canvas.appendChild(textElem);
        this.addToCommonRender(textElem);
    }
    moveToXY(elem, x, y, speed = 10) {
        if (this.html5) {
            if (x == elem.x && y === elem.y)
                return;
            elem.x += elem.vx * this.currentTime;
            elem.y += elem.vy * this.currentTime;
        }
        else {
            var left = parseInt(elem.style.left, 10),
                top = parseInt(elem.style.top, 10),
                dx = left - x,
                dy = top - y,
                i = 1,
                dist = Math.floor(Math.sqrt((x - left) ** 2 + (y - top) ** 2)),
                count = Math.ceil(dist / speed),
                count = 20,
                delay = speed;

            function loop() {
                if (i >= count) { return; }
                i += 1;
                elem.style.left = (left - (dx * i / count)).toFixed(0) + 'px';
                elem.style.top = (top - (dy * i / count)).toFixed(0) + 'px';
                setTimeout(loop, delay);
            }

            loop();
        }

    }


    lazyUpdate(object) {
        object.style.top = parseInt(object.style.top) + 'px';
        object.style.left = parseInt(object.style.left) + 'px';
    }

    addImage(sprite_name, url = null, x, y, callback) {
        if (url) {
        }


    }

    checkWorldBounds() { }

}